/*
**   rl_animation->cpp
**
**   Copyright 2021 Guido Mersmann <geit@geit.de>
**   All rights reserved.
**
*/

/***************************************************************************/

#include "rl_animation.h"
#include "rl_render.h"

/***************************************************************************/

static uint32_t LEDSPERSTRIP[] = { HARDWARE_SEGMENT_1_LEDS, HARDWARE_SEGMENT_2_LEDS, 0 };

struct RL_Animation      animation[ HARDWARE_NUMBEROF_LEDSTRIPS ];
struct RL_AnimationFader fader;

/***************************************************************************/

/* /// Animation_Setup() 
**
*/

/***************************************************************************/

void Animation_Setup( void )
{
int i, num, ledoffset = 0;
struct RL_Animation *tanimation;

	debug(( "\nAnimation_Setup()\n" ));

	memset( &fader, 0, sizeof( struct RL_AnimationFader ) );                                    /* clear entire structure */
	fader.Flags = AFF_FADEOUT|AFF_ACTIVE;                                                      /* make the animation fade in on start */

	for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
		tanimation = &animation[ num ]; /* get animation structure */

		memset( tanimation, 0, sizeof( struct RL_Animation ) );                                /* clear entire structure */

		tanimation->AnimationMode = (uint32_t*) (((uint32_t) (&rlc.rlc_AnimationMode1)) + ( sizeof( uint32_t *) * num ));   /* store config pointer for this strip */

		tanimation->NumberOfLEDs  = LEDSPERSTRIP[ num ];                                       /* setup leds for this strip */
		tanimation->RenderBuffer  = &renderbuffer[ ledoffset ];                                /* setup animation buffer */
		ledoffset += LEDSPERSTRIP[ num ];                                                      /* skip behind current strip */

		for( i = 0 ; i < RACE_NUMBEROF ; i++ ) {                                               /* setup race defaults */
			tanimation->RaceSpeedInit[ i ] = i + 2;
			tanimation->RaceSpeed[ i ]     = i + 2;
			tanimation->RacePos[ i ]       = random( 0, tanimation->NumberOfLEDs - 1 );
		}
	}
	Animation_Randomizer();                                                                /* setup random mode, just in case random mode is used */
}
/* \\\ */

/* /// Animation_Randomizer()
**
*/

/***************************************************************************/

void Animation_Randomizer( void )
{
uint32_t i, rnd, num, syncmode, syncenable = 0;
struct RL_Animation *tanimation;

	debug(( "\nAnimation_Randomizer()\n" ));

/* remove random from bit field, as it makes no sense */
	rlc.rlc_AnimationModeRandom &= ~( ( 1 << ANIMATION_RANDOM ) || ( 1 << ANIMATION_RANDOMSYNC ) );

/* no random mode, then reset to defaults */
	if( !rlc.rlc_AnimationModeRandom ) { /* no random modes defined? -> */
		debug(( "Animation_Randomizer() - No random preset -> using defaults\n" ));
		rlc.rlc_AnimationModeRandom = ANIMATION_RANDOM_MODEFLAGS; /* reset to defaults */
	}

	for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
		tanimation = &animation[ num ]; /* get animation structure */

/* if current mode is the only mode, then there is no need to change */
		rnd = rlc.rlc_AnimationModeRandom;

		if( !( rnd &= ~( 1 << tanimation->BufferedRandomMode ) ) ) {
			debug(( "Animation_Randomizer() - Current mode (%d) is only mode, so no change\n", tanimation->BufferedRandomMode ));
			return; /* only this mode is available, so no change */
		}

/* we use rnd, which lacks the current mode, so the same mode cannot happen twice in a row. */
		while( 1 ) {
			i = random( 0, ANIMATION_LAST - 1 ); /* get random mode */
			if( i != tanimation->BufferedMode && ( rnd & ( 1 << i ) ) && ( i != ANIMATION_RANDOM ) && ( i != ANIMATION_RANDOMSYNC ) ) {
				break;
			}
		}
		tanimation->BufferedRandomMode = i; /* we change the sub mode, so fade will notice */
		/* we have a RANDOMSYNC, so obtain mode data (it does not matter which one)  */
		if( !syncenable && ( *tanimation->AnimationMode == ANIMATION_RANDOMSYNC ) ) {
			syncenable++;
			syncmode = i; /* keep this mode to later write it to all random mode buffers */
		}

		debug(( "Animation_Randomizer() - new mode is (%d)\n", tanimation->BufferedRandomMode ));
	}

/* now we need to handle the random sync mode, if needed. 
** basically if we found a ANIMATION_RANDOMSYNC mode above, we clone
** the generated mode to all animations buffered random mode, so
** the effects restart in sync and equally
*/
	if( syncenable ) {
		for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
			animation[ num ].BufferedRandomMode = syncmode;    /* it does not matter if the mode is actually non random, as this is then never used */
			debug(( "Animation_Randomizer() - RandomSync Enabled: %d\n", syncmode ));
		}
	}

	debug(( "\nAnimation_Randomizer() - Done\n" ));

}
/* \\\ Animation_Randomizer() */

/* /// Animation_Fade() 
**
*/

/***************************************************************************/

uint32_t Animation_GetBrightness( void ) 
{
	return( fader.FadeBrightness );
}
/* \\\  Animation_Fade() */

/* /// Animation_Animate() 
**
*/

/***************************************************************************/

void Animation_Animate( void )
{
uint32_t i, num;
CRGB colorf;
CRGB colorb;
CRGB colort;
struct RL_Animation *tanimation;

//	debug(( "\nAnimation()\n" ));

/* the any animation changes we fade in and out. The process is state driven */
	switch( fader.Flags & ( AFF_ACTIVE|AFF_MODEMASK ) ) { 
		case AFF_FADEOUT|AFF_ACTIVE:  /* fade in animation */
			fader.FadeBrightness  -= ANIMATION_FADESTEP;
			if( fader.FadeBrightness <= 0 ) {
				fader.FadeBrightness = 0;
				fader.Flags &= ~(AFF_ACTIVE); /* fade out done */
			}
			break;
		case AFF_FADEOUT: /* fade out animation is done */
			fader.Flags           &= ~(AFF_MODEMASK); /* disable all modes */
			fader.Flags           |= AFF_ACTIVE|AFF_FADEDELAY; /* activate fade delay */
			fader.FadeDelay        = ANIMATION_FADEDELAY; /* setup delay */
			break;
		case AFF_FADEDELAY|AFF_ACTIVE:  /* fade delay */
			if( fader.FadeDelay ) {
				fader.FadeDelay--;
			} else {
				fader.Flags       &= ~(AFF_ACTIVE); /* fade delay done */
			}
			break;
		case AFF_FADEDELAY: /* fade delay is done */
			fader.Flags           &= ~(AFF_MODEMASK); /* disable all modes */
			fader.Flags           |= AFF_ACTIVE|AFF_FADEIN; /* enable fade in  */
			fader.FadeBrightness   = 0; /* should be obsolete */
			for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
				tanimation = &animation[ num ];
				switch( *tanimation->AnimationMode ) {
#if HARDWARE_NUMBEROF_LEDSTRIPS > 1 
					case ANIMATION_RANDOMSYNC:
#endif
					case ANIMATION_RANDOM:
						tanimation->BufferedMode = tanimation->BufferedRandomMode;  /* set new mode */
						break;
					default:
						tanimation->BufferedMode = *tanimation->AnimationMode;      /* set new mode */
						break;
				}
			/* reset all animation related data here */
				tanimation->Phase            = 0;  /* reset animation phase */
				tanimation->InternalCount1   = 0;
			}
			break;
		case AFF_FADEIN|AFF_ACTIVE:  /* fade in animation */
			fader.FadeBrightness  += ANIMATION_FADESTEP;
			if( fader.FadeBrightness >= 100 ) {
				fader.FadeBrightness = 100;
				fader.Flags &= ~(AFF_ACTIVE); /* fade in done */
			}
			break;
		case AFF_FADEIN:  /* fade in animation is done */
			fader.Flags           &= ~(AFF_MODEMASK); /* disable all modes */
			fader.Flags           |= AFF_ACTIVE; /* activate plain display mode */
			break;
		default:
		case AFF_ACTIVE: /* we run in normal animation mode */
			for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
				tanimation = &animation[ num ];
				switch( *tanimation->AnimationMode ) {
					case ANIMATION_RANDOM:
#if HARDWARE_NUMBEROF_LEDSTRIPS > 1 
					case ANIMATION_RANDOMSYNC:
#endif
						if( tanimation->BufferedRandomMode != tanimation->BufferedMode ) {
							fader.Flags          |= AFF_ACTIVE|AFF_FADEOUT;         /* turn animation on again. This time "out" mode */
							fader.FadeBrightness  = 100;                            /* should be obsolete */
						}
						break;
					default:
						if( *tanimation->AnimationMode != tanimation->BufferedMode ) {
							fader.Flags          |= AFF_ACTIVE|AFF_FADEOUT;         /* turn animation on again. This time "out" mode */
							fader.FadeBrightness  = 100;                            /* should be obsolete */
						}
						break;
				}
			}
			break;
	}
	for( num = 0 ; num < HARDWARE_NUMBEROF_LEDSTRIPS ; num++ ) {
		tanimation = &animation[ num ];

		/* pre calculate color based on brightness */
		colorf = rlc.rlc_RGBColor;
		colorb = rlc.rlc_RGBColorAlternate;

		switch( tanimation->BufferedMode ) {
/* /// STATIC */
			case ANIMATION_STATIC:
				if( num ) {
					colorf = colorb;
				}
				for( i = 0 ; i < tanimation->NumberOfLEDs ; i++ ) {
					tanimation->RenderBuffer[ i ] = colorf;
				}
				break;
/* \\\ STATIC */
/* /// Rainbow */
			case ANIMATION_RAINBOWSPARKLE:
			case ANIMATION_RAINBOW:

				tanimation->Phase++;
				tanimation->Phase &= 0xff;

				fill_rainbow( tanimation->RenderBuffer, tanimation->NumberOfLEDs, tanimation->Phase, 7 );
/* Sparkle mode is just the normal rainbow with random sparkles */
				if( tanimation->BufferedMode == ANIMATION_RAINBOWSPARKLE ) {
					if( random8(100) < RAINBOW_SPARKLERATE ) {
						tanimation->RenderBuffer[ random16( tanimation->NumberOfLEDs - 1  ) ] = CRGB( 255, 255, 255 );
					}
				}
				break;
/* \\\ Rainbow */
/* /// FIRE */
			case ANIMATION_FIREPLACE:
			case ANIMATION_BLUEFIRE:
			case ANIMATION_GREENFIRE:
/* handle fire animation speed */
				if( tanimation->Phase ) {
					tanimation->Phase--;
					break;
				}
				tanimation->Phase = FIRE_SPEED; /* reset speed */

				{ 
					byte *fireheat;
					int   firesize;
					int   i;
/* pick the right heat chamber depending on the ring used */
					firesize = tanimation->NumberOfLEDs / 2;   /* we mirror the fire, so only 50% of the ring is used as height */
					fireheat = &tanimation->Heat[0];           /* heat buffer */
					/* cooling phase */
					for( i = 0 ; i < firesize ; i++ ) {
						fireheat[ i ] = max( 0L, fireheat[ i ] - random( 0, ( ( FIRE_COOLING * 10 ) / firesize ) + 2 ));
					}
					/* drift and defuse */
					for( i = firesize - 1 ; i >= 3 ; i-- ) {
						fireheat[ i ] = ( 
								fireheat[ ( i - 0 ) ] * FIRE_BLENDSELF      + 
								fireheat[ ( i - 1 ) ] * FIRE_BLENDNEIGHBOR1 +
								fireheat[ ( i - 2 ) ] * FIRE_BLENDNEIGHBOR2 +
								fireheat[ ( i - 3 ) ] * FIRE_BLENDNEIGHBOR3 )
								/ FIRE_BLENDDIVIDER;
					}
					for( i = 0 ; i < FIRE_SPARKS ; i++ ) {
						if( random(255) < FIRE_SPARKING ) {
							int sparkoffset = random( FIRE_SPARKHEIGHT );
							fireheat[ sparkoffset ] = fireheat[ sparkoffset ] + random( 160, 255 );
						}
					}
					/* heat to color */
					for( i = 0 ; i < firesize ; i++ ) {
						byte gun;
						colorf = HeatColor( fireheat[ i ] );
						switch( tanimation->BufferedMode ) {
							case ANIMATION_BLUEFIRE:      /* swap the red gun with the blue to create blue fure */
								gun = colorf.b;
								colorf.b = colorf.r;
								colorf.r = gun;
								break;
							case ANIMATION_GREENFIRE:    /* swap the red gun with the green to create green fire */
								gun = colorf.g;
								colorf.g = colorf.r;
								colorf.r = gun;
								break;
						}
						tanimation->RenderBuffer[ i                    ] = colorf;   /* store color value */
						tanimation->RenderBuffer[ 2 * firesize - i - 1 ] = colorf;   /* mirror the same color to the other side */
					}
				}
				break;
/* \\\ FIRE */
/* /// RGBFade */
			case ANIMATION_RGBFADE:
				{ uint8_t gun;
					CRGB tcolor;
					gun = tanimation->Phase & 0xff; /* animation->Phase is our color gun value */
					if( tanimation->Phase >= 256 ) { /* if phase is bigger than 255, then we reverse fading */
						gun = 255 - gun;
					}
					/* select a color */
					switch( ( tanimation->InternalCount1 % 3 ) ) {
						case 0:
							tcolor = CRGB( gun, 0, 0 );
							break;
						case 1:
							tcolor = CRGB( 0, gun, 0 );
							break;
						case 2:
							tcolor = CRGB( 0, 0, gun );
							break;
					}
					/* write to renderbuffer */
					for( i = 0 ; i < tanimation->NumberOfLEDs ; i++ ) {
						tanimation->RenderBuffer[ i ] = tcolor;
					}
					tanimation->Phase++;
					if( ( tanimation->Phase == 512 ) ) { /* two cycles, so change color */
						tanimation->InternalCount1++; /* select next r, g or gun */
						tanimation->Phase = 0;
					}
				}
				break;
/* \\\ RGBFade */
/* /// TWINKLE */
			case ANIMATION_TWINKLE:
				fadeToBlackBy( tanimation->RenderBuffer, tanimation->NumberOfLEDs, TWINKLE_FADEAWAY );
				tanimation->RenderBuffer[ random( tanimation->NumberOfLEDs ) % tanimation->NumberOfLEDs ] = CRGB( random( 255 ), random( 255 ), random( 255 ) );
				break;
/* \\\ TWINKLE */
/* /// RACE */
			case ANIMATION_RACE:
				{
				uint32_t hue = 0;
					/* fade previous set pixels towards black to create a trail */
					fadeToBlackBy( tanimation->RenderBuffer, tanimation->NumberOfLEDs, RACE_TRAIL_FADEAWAY );
					for( i = 0 ; i < RACE_NUMBEROF ; i++ ) {
						/* now we use a beatsine with different bpm for each lifeform to randomly move the head pixel */
						tanimation->RenderBuffer[ beatsin16( i + 7, 0, tanimation->NumberOfLEDs - 1 ) ] += CHSV( hue, 200, 255 );
						hue += 256 / RACE_NUMBEROF; /* spread color equaly on the color palette depending on the amount of racers */
					}
				}
				break;
/* \\\ RACE */
/* /// RACE (LOOP) */
			case ANIMATION_RACELOOP:
				{
				uint32_t hue = 0;
					/* fade previous set pixels towards black to create a trail */
					fadeToBlackBy( tanimation->RenderBuffer, tanimation->NumberOfLEDs, RACE_TRAIL_FADEAWAY );
					for( i = 0 ; i < RACE_NUMBEROF ; i++ ) {
						tanimation->RaceSpeed[ i ] = ( tanimation->RaceSpeed[ i ] + 1 ) % tanimation->RaceSpeedInit[ i ];
						if( !tanimation->RaceSpeed[ i ] ) {
							tanimation->RacePos[ i ] = ( tanimation->RacePos[ i ] + 1 ) % ( tanimation->NumberOfLEDs );
						}
						tanimation->RenderBuffer[ tanimation->RacePos[ i ] ] += CHSV( hue, 200, 255 );
						hue += 256 / RACE_NUMBEROF; /* spread color equaly on the color palette depending on the amount of racers */
					}
				}
				break;
/* \\\ RACE (LOOP) */
/* /// COLORCOMET STATIC */
			case ANIMATION_COLORCOMETSTATIC:
				fadeToBlackBy( tanimation->RenderBuffer, tanimation->NumberOfLEDs, COLORCOMET_TRAIL_FADEAWAY );
				tanimation->RenderBuffer[ beatsin16( COLORCOMET_BPM, 0, tanimation->NumberOfLEDs - 1 ) ] += colorf;
				break;
/* \\\ COLORCOMET STATIC */
/* /// COLORCOMET */
			case ANIMATION_COLORCOMET:
				fadeToBlackBy( tanimation->RenderBuffer, tanimation->NumberOfLEDs, COLORCOMET_TRAIL_FADEAWAY );
				tanimation->RenderBuffer[ beatsin16( COLORCOMET_BPM, 0, tanimation->NumberOfLEDs - 1 ) ] += CHSV( tanimation->InternalCount1++, 255, 192);
				break;
/* \\\ COLORCOMET */

		} /* switch */
	} /* for( num ) */

//	debug(( "\nAnimation() - Done\n" ));

}
/* \\\ */
