/*
**    rl_button.cpp
**
**   Copyright 2021 Guido Mersmann <geit@geit.de>
**   All rights reserved.
**
*/

/***************************************************************************/

#define DISABLE_DEBUG     1  /* this overrides the global debug */

/***************************************************************************/

#include "main.h"
#include "rl_button.h"
#include "rl_global.h"
#include "rl_led.h"
#include "rl_wifi.h"

/***************************************************************************/

/*
** disable debug locally if requested
*/

#if DISABLE_DEBUG
#undef  debug
#define debug(x)
#endif

/***************************************************************************/

static uint32_t      buttoncount;
static uint32_t      buttonstate;
static unsigned long millislast;

/***************************************************************************/

#define PRESSED                     LOW   /* more readble code */

/*
** button press state engine
*/

enum{
	BUTTON_STATE_COOLDOWN      = 0,
	BUTTON_STATE_WAITFORRELEASE,
	BUTTON_STATE_WAITFORPRESS,
};

/***************************************************************************/

/* /// Button_Setup()
**
*/

/***************************************************************************/

void Button_Setup( void )
{
	debug(( "\nButton_Setup()\n" ));

	pinMode( HARDWARE_BUTTON_PIN, INPUT_PULLUP );
	pinMode( HARDWARE_ONBOARDLED_PIN, OUTPUT );

	if( digitalRead( HARDWARE_BUTTON_PIN ) == PRESSED ) {
		debuglimit( 9, debug(( "\nButton_Setup() - PowerOnPress: performing settings reset, settings save and reboot\n" )) );
		/* we have a power on press */
		Global_SettingsReset();   /* factory reset settings */
		Global_SettingsSave();    /* write settings to flash */
		LED_BlinkRythm( LED_NOTIFICATION_DURATION, LED_NOTIFICATION_BLINKSPEED, LED_NOTIFICATION_BLINKSPEED, BUTTON_RYTHM_SHORTLONG );   /* signal user to release the button */
		ESP.restart();
	}

	millislast    = millis();               /* store current millis */
	buttonstate   = BUTTON_STATE_COOLDOWN;  /* we start with a cool down */

}
/* \\\ Button_Setup() */
/* /// Button_Handle()
**
*/

/***************************************************************************/

void Button_Handle( void )
{
bool          state      = digitalRead( HARDWARE_BUTTON_PIN );
unsigned long millistemp = millis();

	debugex(( "\nButton_Handle()\n" ));

	switch( buttonstate ) {
		case BUTTON_STATE_COOLDOWN:                           /* we wait for the button to be released for some time */
			if( state != PRESSED ) {                          /* button released, then check cool down periode */
				if( ( millistemp - millislast ) >= BUTTON_THRESSHOLD_COOLDOWN ) {
					debuglimit( 1, debug(( "Button_Handle() - Cool down done\n" )) );
					buttoncount = 0;                          /* reset press counter */
					millislast  = millistemp;                 /* update timestamp */
					buttonstate = BUTTON_STATE_WAITFORPRESS;  /* we wait for a press now */
				}
			} else { 
				millislast  = millistemp;                     /* button still pressed: reset cool down */
				debuglimit( 2, debug(( "Button_Handle() - Cool down in progress\n" )) );
			}
			break;

		case BUTTON_STATE_WAITFORPRESS:                       /* button is not pressed for a long time (post cool down) */
			if( state != PRESSED )  {                         /* still released? */
				debuglimit( 3, debug(( "Button_Handle() - BUTTON_STATE_WAITFORPRESS: still not pressed\n" )) );
				if( buttoncount ) {                           /* only check for action if there actually was a previous button press */
					if( ( millistemp - millislast ) >= BUTTON_THRESSHOLD_RELEASE ) {
						debug(( "Button_Handle() - BUTTON_STATE_WAITFORPRESS: Click timeout: Action # is %d\n", buttoncount ));
						switch( buttoncount ) {
							/* insert short press actions here */
							case 1:
								rlrd.rlrd_GlobalMode = RLRD_GLOBALMODE_ALWAYSON;  /* turn our ring back to auto mode */
								break;
							case 2:
								rlrd.rlrd_GlobalMode = RLRD_GLOBALMODE_AUTO;  /* turn our ring back to auto mode */
								break;
#ifndef RELEASE_FIRMWARE_BIN
							case 4:
								Wifi_SwapNetwork();
								break;
#endif
						}
						millislast  = millistemp;                 /* make sure cooldown starts now */
						buttonstate = BUTTON_STATE_COOLDOWN;      /* button needs to be released for some time before we continue */
					}
				}
			} else {
				millislast  = millistemp;                     /* make sure new state gets the right timing */
				buttonstate = BUTTON_STATE_WAITFORRELEASE;    /* we now wait for release */
				buttoncount++;                                /* we got an external button press, so make it count */
				debuglimit( 4, debug(( "Button_Handle() - BUTTON_STATE_WAITFORPRESS: pressed (%d)\n", buttoncount )) );
			}
			break;

		case BUTTON_STATE_WAITFORRELEASE:
			if( state == PRESSED )  {                         /* still pressed ? */
															  /* still pressed, so handle long press */
				debuglimit( 5, debug(( "Button_Handle() - BUTTON_STATE_WAITFORRELEASE: still pressed (%d)\n", buttoncount )) );
				if( ( millistemp - millislast ) >= BUTTON_THRESSHOLD_LONGPRESS ) {
					debug(( "\nButton_Handle() - BUTTON_STATE_WAITFORRELEASE: Longpress: Action # is %d\n", buttoncount ));
					switch( buttoncount ) {
						/* insert long press actions here */
						case 1:
							rlrd.rlrd_GlobalMode = RLRD_GLOBALMODE_ALWAYSOFF;  /* turn our ring back to auto mode */
							break;
						case 4:                /* four presses and hold the last one */
							Global_SettingsReset();   /* factory reset settings */
							Global_SettingsSave();    /* write settings to flash */
							LED_BlinkRythm( LED_NOTIFICATION_DURATION, LED_NOTIFICATION_BLINKSPEED, LED_NOTIFICATION_BLINKSPEED, BUTTON_RYTHM_SHORTLONG );   /* signal user to release the button */
							ESP.restart();
					}
					millislast  = millistemp;                 /* make sure cooldown starts now */
					buttonstate = BUTTON_STATE_COOLDOWN;      /* button needs to be released for some time before we continue */
					break;
				}
			} else {
				debuglimit( 6, debug(( "Button_Handle() - BUTTON_STATE_WAITFORRELEASE: released (%d)\n", buttoncount )) );
				millislast  = millistemp;                     /* make sure cooldown starts now */
				buttonstate = BUTTON_STATE_WAITFORPRESS;
				break;
			}
			break;
	}
}
/* \\\ */

/***************************************************************************/
