/*
**   rl_display.cpp
**
**   Copyright 2021 Guido Mersmann <geit@geit.de>
**   All rights reserved.
**
*/

/***************************************************************************/

#include "main.h"
#include "rl_display.h"
#include "rl_global.h"
#include "rl_render.h"
#include "rl_time.h"

/***************************************************************************/

CRGB ledfield[ DISPLAY_TOTAL_LEDS ]; /* the real LED buffer */

/***************************************************************************/

/*
** These are the color correction modes supported by FastLED
**
** NOTE: If you insert, remove or swap entries here, you need
** to adapt the matching TAB_CYLE entry in rl_wifi.cpp, or the
** web interface wont match the modes anymore.
*/

CRGB TAB_LEDColorCorrection[] PROGMEM = {
	UncorrectedColor,
	TypicalSMD5050,
	TypicalLEDStrip,
	Typical8mmPixel,
	TypicalPixelString
};

/***************************************************************************/

/*
** These are the color temperature correction modes supported by FastLED
**
** NOTE: If you insert, remove or swap entries here, you need
** to adapt the matching TAB_CYLE entry in rl_wifi.cpp, or the
** web interface wont match the modes anymore.
*/

CRGB TAB_LEDColorTemperature[] PROGMEM = {
	UncorrectedTemperature,
	Candle,
	Tungsten40W,
	Tungsten100W,
	Halogen,
	CarbonArc,
	HighNoonSun,
	DirectSunlight,
	OvercastSky,
	ClearBlueSky,
	WarmFluorescent,
	StandardFluorescent,
	CoolWhiteFluorescent,
	FullSpectrumFluorescent,
	GrowLightFluorescent,
	BlackLightFluorescent,
	MercuryVapor,
	SodiumVapor,
	MetalHalide,
	HighPressureSodium
};

/* /// Display_UpdateSettings()
**
*/

/***************************************************************************/

void Display_UpdateSettings( void )
{

	FastLED.setCorrection( TAB_LEDColorCorrection[ rlc.rlc_LEDColorCorrection ] );
	FastLED.setTemperature( TAB_LEDColorTemperature[ rlc.rlc_LEDColorTemperature ] );

}
/* \\\ */

/* /// Display_Startup() 
**
*/

/***************************************************************************/

void Display_Startup( void )
{
	debug(( "\nDisplay_Startup()\n"));

/* set brightness to zero */

	FastLED.setBrightness( 0 );

/* protect our power supply from burning the house down */

	FastLED.setMaxPowerInVoltsAndMilliamps( PSU_VOLTAGE, PSU_MILLIAMPS );

/* for now we set entire clock to black as default */

	fill_solid( ledfield, DISPLAY_TOTAL_LEDS, CRGB::Black);

/* predefine our segments containing LEDstrip */

	FastLED.addLeds<HARDWARE_SEGMENT_1_LEDTYPE, HARDWARE_SEGMENT_1_PIN, GRB>( &ledfield[ 0 ]                      , 0, HARDWARE_SEGMENT_1_LEDS );
#if HARDWARE_NUMBEROF_LEDSTRIPS > 1 
	FastLED.addLeds<HARDWARE_SEGMENT_2_LEDTYPE, HARDWARE_SEGMENT_2_PIN, GRB>( &ledfield[ HARDWARE_SEGMENT_1_LEDS ], 0, HARDWARE_SEGMENT_2_LEDS );
#endif

	FastLED.show();

/* enable color correction */

	Display_UpdateSettings();

/* clear our render buffer */

	Render_ClearBuffer();

/* set brightness to max */

	FastLED.setBrightness( 255 * DISPLAY_BRIGHTNESS_MIN / 100 );

	debug(( "LEDS: %d\n", DISPLAY_TOTAL_LEDS ));

}
/* \\\ */

/* /// Display_Handle()
**
*/

/***************************************************************************/
void Display_Handle( void )
{

	show_at_max_brightness_for_power();

}
/* \\\ */
/* /// Display_Update()
**
*/

/***************************************************************************/

void Display_Update( void )
{
uint32_t i;

	for( i = 0 ; i < DISPLAY_TOTAL_LEDS ; i++ ) {
		ledfield[ i ] = renderbuffer[ i ];
	}

/* status LED */

	if( rlc.rlc_DisplayStatusLEDMode != DISPLAY_STATUSLEDMODE_OFF ) { /* when enabled override color */
		ledfield[ RINGLIGHT_STATUSLED_NUM ] = rlrd.rlrd_StatusLED;   /* set current LED status color */
	}

/* compose brightness */

	Global_UpdateDisplayBrightness();

/* update LEDs */

	show_at_max_brightness_for_power();
}
/* \\\ */

/* /// Display_Clear()
**
*/

/***************************************************************************/

void Display_Clear( void )
{
int i;

	debugex(( "\nDisplay_Clear()"));

	for( i = 0 ; i < DISPLAY_TOTAL_LEDS ; i++ ) {
		ledfield[ i ] = CRGB::Black;
	}

	/* update LEDs */

	show_at_max_brightness_for_power();
}
/* \\\ */
/* /// Display_SetStatusLED()
**
*/

/***************************************************************************/

void Display_SetStatusLED( uint32_t mode, CRGB color )
{
	debugex(( "\nDisplay_SetStatusLED()"));

	if( rlc.rlc_DisplayStatusLEDMode == mode ) {   /* only change LED when mode is right */
		rlrd.rlrd_StatusLED = color;
	}
}
/* \\\ Display_SetStatusLED() */

/***************************************************************************/
