/*
**   rl_webserver.cpp
**
**   Copyright 2021 Guido Mersmann <geit@geit.de>
**   All rights reserved.
**
*/

/***************************************************************************/

#include <WiFi.h>
#include <WebServer.h>

#include "main.h"
#include "rl_fauxmo.h"
#include "rl_global.h"
#include "rl_html.h"
#include "rl_render.h"
#include "rl_sound.h"
#include "rl_testmode.h"
#include "rl_webserver.h"
#include "rl_wifi.h"

/***************************************************************************/

#define WEBSERVER_COOKIENAME    APPLICATIONNAME_NSUC "SESSIONID"

/***************************************************************************/

/*
** the wifi action defines
*/

enum {
	ACTION_NONE = 0,
	ACTION_SETTINGSLOAD,
	ACTION_SETTINGSSAVE,
	ACTION_SETTINGSRESET,
	ACTION_SWAPNETWORK,
	ACTION_SYSTEMREBOOT,
	ACTION_SOUNDTEST1,
	ACTION_SOUNDTEST2,
	ACTION_WIFISAVEREBOOT,
	ACTION_CHANGEPASSWORD,
	ACTION_SAVEFAUXMO
};

WebServer server( WEBSERVER_PORT );

uint32_t webserver_action       = ACTION_NONE;

String htmlpage = "";         /* web page creation is using this as storage */

/***************************************************************************/

/*
** navigation table
*/

#define PAGEID_STATUS               1
#define PAGEID_VISUAL               2
#define PAGEID_SOUND                4
#define PAGEID_GLOBAL               8
#define PAGEID_MISC                16
#define PAGEID_WIFI                32
/* the following are not in TOPNAV */
#define PAGEID_LOGIN               64
#define PAGEID_PASS               128

struct pageiddata {
	const uint16_t p_PageID;
	const char    *p_URI;
	const char    *p_Name;
};

static const struct pageiddata TAB_PAGEIDDATA[] PROGMEM = {
	{ PAGEID_STATUS             , "/status" , "Status"  },
	{ PAGEID_VISUAL             , "/visual" , "Visual"  },
	{ PAGEID_SOUND              , "/sound"  , "Sound"   },
	{ PAGEID_GLOBAL             , "/global" , "Global"  },
	{ PAGEID_MISC               , "/misc"   , "Misc"    },
	{ PAGEID_WIFI|PAGEID_PASS   , "/wifi"   , "Network" },
	{ 0                         , NULL      , NULL      }
};

#define PAGEIDDATA_TOPNAV_END   6

/***************************************************************************/

/*
** available animation modes
*/

static const char *TAB_CYCLE_ANIMATIONMODE[] PROGMEM = {
	"Static",                /* make sure this table corressponds to the defines in animation.h */
	"Rainbow",
	"Rainbow Sparkle",
	"Fire Place",
	"Blue Fire",
	"Green Fire",
	"RGB Fade",
	"Twinkle",
	"Race",
	"Race (Loop)",
	"Color Comet",
	"Color Comet (Static)",
	"Random",
#if HARDWARE_NUMBEROF_LEDSTRIPS > 1 
	"Random (sync)",
#endif
	NULL
};

/***************************************************************************/

/*
** available testmodes
*/
#if ENABLE_TESTMODE
static const char *TAB_CYCLE_TESTMODE[] PROGMEM = {
	"Off",                   /* make sure this table corressponds to the defines in testmode.h */
	"Test",                  /* high change of crashing. used to test new code */
	"Auto",
	"Train",
	"Red",
	"Green",
	"Blue",
	"White",
	"StripLength",
	"StripLength-1",
	NULL
};
#endif

/***************************************************************************/

/*
** global modes
*/

static const char *TAB_CYCLE_GLOBALMODE[] PROGMEM = {
	"Auto",                  /* display gets dark, no animation or other processing beside second OTA/Web core */
	"Always Off",
	"Always On",
#if ENABLE_TESTMODE
	"Testmode",              /* must be the last entry to avoid a config mess */
#endif
	NULL
};

/***************************************************************************/

/*
** status LED modes
*/

static const char *TAB_CYCLE_STATUSLEDMODE[] PROGMEM = {
	"Off",                   /* LED acts as normal device LED */
	"WiFi Status",           /* LED shows wifi status */
	NULL
};

/***************************************************************************/

/*
** LED color correction name defines
*/

static const char *TAB_CYCLE_LEDCOLORCORRECTION[] PROGMEM = {
	"Uncorrected Color",
	"Typical SMD5050",
	"Typical LED Strip",
	"Typical 8mm Pixel",
	"Typical Pixel String",
	NULL
};

/***************************************************************************/

/*
** LED color temperature name defines
*/

static const char *TAB_CYCLE_LEDCOLORTEMPERATURE[] PROGMEM = {
	"Uncorrected Temperature",
	"Candle",
	"Tungsten 40W",
	"Tungsten 100W",
	"Halogen",
	"Carbon Arc",
	"High Noon Sun",
	"Direct Sunlight",
	"Overcast Sky",
	"Clear BlueSky",
	"Warm Fluorescent",
	"Standard Fluorescent",
	"CoolWhite Fluorescent",
	"Full Spectrum Fluorescent",
	"Grow Light Fluorescent",
	"Black Light Fluorescent",
	"Mercury Vapor",
	"Sodium Vapor",
	"Metal Halide",
	"High Pressure Sodium",
	NULL
};

/***************************************************************************/

/*
** name array defining all sample within the 
** SD-Card of the DFPlayer sound module.
*/

static const char *TAB_CYCLE_SOUNDS[] PROGMEM = {
	"off",
	"Alarm 1",
	"Alarm 2",
	"Alarm 3",
	"Alarm 4",
	"Alarm 5",
	"Alarm 6",
	"Alarm 7",
	"Alarm 8",
	"Alarm 9",
	"Alarm 10",
	"Alarm 11",
	"Alert 1",
	"Bell 1",
	"Bell 2",
	"Bell 4",
	"Blank",
	"Door 1",
	"Door 2",
	"Door 3",
	"Door 4",
	"Door 5",
	"Notify 1",
	"Notify 2",
	"Notify 3",
	"Notify 4",
	"Notify 5",
	"Notify 6",
	"Phone 1",
	"Startup 2",
	"Test 1",
	"Test 2",
	NULL
};

/***************************************************************************/

/* /// HTML PAGE DATA
**
*/

/***************************************************************************/

char temppass[ WEBSERVER_PASSWORD_SIZEOF + 2 ];

/***************************************************************************/

static struct ws_Settings TAB_SETTINGS[] PROGMEM = {
	{ PAGEID_VISUAL , WSTYPE_TITLE   ,            0, NULL                          , "Visual Configuration"    , NULL                                       , 0                                           ,   0  ,   0, NULL },
#if HARDWARE_NUMBEROF_LEDSTRIPS > 1 
	{ PAGEID_VISUAL , WSTYPE_CYCLE   ,            0, "GenAnimatio1ID"              , "Mode Inner:"             , &rlc.rlc_AnimationMode1                    , (uint32_t) &TAB_CYCLE_ANIMATIONMODE[0]      ,   0  ,   0, NULL },
	{ PAGEID_VISUAL , WSTYPE_CYCLE   ,            0, "GenAnimatio2ID"              , "Mode Outer:"             , &rlc.rlc_AnimationMode2                    , (uint32_t) &TAB_CYCLE_ANIMATIONMODE[0]      ,   0  ,   0, NULL },
#else
	{ PAGEID_VISUAL , WSTYPE_CYCLE   ,            0, "GenAnimatio1ID"              , "Mode:"                   , &rlc.rlc_AnimationMode1                    , (uint32_t) &TAB_CYCLE_ANIMATIONMODE[0]      ,   0  ,   0, NULL },
#endif

	{ PAGEID_VISUAL , WSTYPE_REFFLAG ,            0, "FavoriteID"                  , "Favorite"                , &rlc.rlc_AnimationModeRandom               , (uint32_t) &rlc.rlc_AnimationMode1          ,   0  ,   0, "onclick=\"CB('%s')\"" },
	{ PAGEID_VISUAL , WSTYPE_SLIDER  ,            0, "GenBrightnessID"             , "Brightness:"             , &rlc.rlc_Brightness                        , 0                                           , 100  ,   0, NULL },
	{ PAGEID_VISUAL , WSTYPE_LABEL   , WSF_SR      , NULL                          , "Color:"                  , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_VISUAL , WSTYPE_COLOR   , WSF_SRT     , "GenFaceID"                   , NULL                      , (uint32_t*) &rlc.rlc_RGBColor              , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_VISUAL , WSTYPE_COLOR   ,            0, "GenAlternateID"              , NULL                      , (uint32_t*) &rlc.rlc_RGBColorAlternate     , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_VISUAL , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },

	{ PAGEID_SOUND  , WSTYPE_TITLE   ,            0, NULL                          , "Sound Configuration"     , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_ONOFF   ,            0, "SoundEnableID"               , "Sound:"                  , &rlc.rlc_Flags                             , RLCF_SOUNDENABLE                            ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_NUMBER  ,            0, "SoundNightStartID"           , "Night Start:"            , &rlc.rlc_SoundVolumeNightStart             , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_SLIDER  ,            0, "SoundVolumeNightID"          , "Volume at night:"        , &rlc.rlc_SoundVolumeNight                  , 0                                           , 100  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_NUMBER  ,            0, "SoundNightEndID"             , "Day Start:"              , &rlc.rlc_SoundVolumeNightEnd               , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_SLIDER  ,            0, "SoundVolumeDayID"            , "Volume at day:"          , &rlc.rlc_SoundVolumeDay                    , 0                                           , 100  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_CYCLE   ,            0, "SoundSelectBootID"           , "Boot Sample:"            , &rlc.rlc_SoundIDBoot                       , (uint32_t) &TAB_CYCLE_SOUNDS[0]             ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_CYCLE   ,            0, "SoundSelectHourID"           , "Hour:"                   , &rlc.rlc_SoundIDHour                       , (uint32_t) &TAB_CYCLE_SOUNDS[0]             ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_CYCLE   ,            0, "SoundSelectHalfHourID"       , "Half Hour:"              , &rlc.rlc_SoundIDHalfHour                   , (uint32_t) &TAB_CYCLE_SOUNDS[0]             ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_TITLE   ,            0, NULL                          , "Test Section"            , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_BUTTON  , WSF_SRT     , "SoundTest1ID"                , "Test1"                   , NULL                                       , ACTION_SOUNDTEST1                           ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_BUTTON  ,            0, "SoundTest2ID"                , "Test2"                   , NULL                                       , ACTION_SOUNDTEST2                           ,   0  ,   0, NULL },
	{ PAGEID_SOUND  , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },

	{ PAGEID_GLOBAL , WSTYPE_TITLE   ,            0, NULL                          , "Global Configuration"    , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_CYCLE   ,            0, "GlobalModeID"                , "Mode:"                   , &rlrd.rlrd_GlobalMode                      , (uint32_t) &TAB_CYCLE_GLOBALMODE[0]         ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_CYCLE   ,            0, "StatusLEDID"                 , "Status LED:"             , &rlc.rlc_DisplayStatusLEDMode              , (uint32_t) &TAB_CYCLE_STATUSLEDMODE[0]      ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_TITLE   ,            0, NULL                          , "Color"                   , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_CYCLE   ,            0, "GlobalLEDCorrID"             , "LED Color Correction:"   , &rlc.rlc_LEDColorCorrection                , (uint32_t) &TAB_CYCLE_LEDCOLORCORRECTION[0] ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_CYCLE   ,            0, "GlobalLEDTempID"             , "LED Color Temperature:"  , &rlc.rlc_LEDColorTemperature               , (uint32_t) &TAB_CYCLE_LEDCOLORTEMPERATURE[0],   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_TITLE   ,            0, NULL                          , "Power Configuration"     , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_NUMBER  ,            0, "GlobalPowerNightEndID"       , "Turn on at:"             , &rlc.rlc_PowerStateNightEnd                , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_NUMBER  ,            0, "GlobalPowerNightStartID"     , "Turn off at:"            , &rlc.rlc_PowerStateNightStart              , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_TITLE   ,            0, NULL                          , "Brightness Configuration", NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_NUMBER  ,            0, "GlobalBrightnessNightStartID", "Night Start:"            , &rlc.rlc_DisplayBrightnessNightStart       , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_SLIDER  ,            0, "GlobalBrightnessNightID"     , "Brightness at night:"    , &rlc.rlc_DisplayBrightnessNight            , 0                                           , 100  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_NUMBER  ,            0, "GlobalBrightnessNightEndID"  , "Day Start:"              , &rlc.rlc_DisplayBrightnessNightEnd         , 0                                           ,  23  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_SLIDER  ,            0, "GlobalBrightnessDayID"       , "Brightness at day:"      , &rlc.rlc_DisplayBrightnessDay              , 0                                           , 100  ,   0, NULL },
	{ PAGEID_GLOBAL , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },

#if ENABLE_TESTMODE
	{ PAGEID_MISC   , WSTYPE_TITLE   ,            0, NULL                          , "Miscellaneus"            , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_CYCLE   ,            0, "TestModeID"                  , "Test Mode:"              , &rlrd.rlrd_TestMode                        , (uint32_t) &TAB_CYCLE_TESTMODE[0]           ,   0  ,   0, NULL },
#endif
	{ PAGEID_MISC   , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_TITLE   ,            0, NULL                          , "Settings"                , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_BUTTON  , WSF_C       , "SettingsSaveID"              , "Save"                    , NULL                                       , ACTION_SETTINGSSAVE                         ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_BUTTON  , WSF_C       , "SettingsLoadID"              , "Load"                    , NULL                                       , ACTION_SETTINGSLOAD                         ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_BUTTON  , WSF_C       , "SettingsDefaultsID"          , "Factory Reset"           , NULL                                       , ACTION_SETTINGSRESET                        ,   0  ,   0, NULL },
#ifndef RELEASE_FIRMWARE_BIN
	{ PAGEID_MISC   , WSTYPE_BUTTON  , WSF_C       , "SettingsAltID"               , "Swap Wifi Settings"      , NULL                                       , ACTION_SWAPNETWORK                          ,   0  ,   0, NULL },
#endif
	{ PAGEID_MISC   , WSTYPE_TITLE   ,            0, NULL                          , "System"                  , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_PLAIN   ,            0, "<td style=\"text-align:center\"><button id=\"bt\" name=\"logout\" value=\"1\" onclick=\"window.location.href='/login?DISCONNECT=YES'\">Logout</button>", NULL, NULL, 0, 0, 0, NULL },
	{ PAGEID_MISC   , WSTYPE_BUTTON  , WSF_C       , "SysID"                       , "Reboot"                  , NULL                                       , ACTION_SYSTEMREBOOT                         ,   0  ,   0, NULL },
	{ PAGEID_MISC   , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },

	{ PAGEID_WIFI   , WSTYPE_TITLE   ,            0, NULL                          , "Wifi Configuration"      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_WIFI   , WSTYPE_TEXT    ,            0, "WIFISSIDID"                  , "SSID:"                   , (uint32_t*) &rlc.rlc_WifiSSID              , WIFI_SSID_SIZEOF                            ,   0  ,   0, NULL },
	{ PAGEID_WIFI   , WSTYPE_PASSWORD,            0, "WIFIPASSID"                  , "Password:"               , (uint32_t*) &rlc.rlc_WifiPass              , WIFI_PASS_SIZEOF                            ,   0  ,   0, NULL },
	{ PAGEID_WIFI   , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_WIFI   , WSTYPE_BUTTON  ,            0, "WIFISaveRebootID"            , "Save and Connect"        , NULL                                       , ACTION_WIFISAVEREBOOT                       ,   0  ,   0, NULL },
	{ PAGEID_WIFI   , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
  
	{ PAGEID_PASS   , WSTYPE_TITLE   ,            0, NULL                          , "Login Configuration"     , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_PASSWORD,            0, "Pass1ID"                     , "Password:"               , NULL                                       , WEBSERVER_PASSWORD_SIZEOF                   ,   0  ,   0, "onchange=\"PW('%s')\"" },
	{ PAGEID_PASS   , WSTYPE_PASSWORD,            0, "Pass2ID"                     , "Repeat:"                 , NULL                                       , WEBSERVER_PASSWORD_SIZEOF                   ,   0  ,   0, "onchange=\"PW('%s')\"" },
	{ PAGEID_PASS   , WSTYPE_PASSWORD, WSF_H       , "SECID"                       , "hidden:"                 , (uint32_t*) &temppass                      , WEBSERVER_PASSWORD_SIZEOF                   ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_LABEL   , WSF_L       , "PassInfoID"                  , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_BUTTON  ,            0, "PASSSAVEID"                  , "Save and Update"         , NULL                                       , ACTION_CHANGEPASSWORD                       ,   0  ,   0, NULL },
#if ENABLE_FAUXMO
	{ PAGEID_PASS   , WSTYPE_TITLE   ,            0, NULL                          , "Alexa Support"           , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_TEXT    ,            0, "FAUXMOID"                    , "Name:"                   , (uint32_t*) &rlc.rlc_FauxmoDeviceName      , FAUXMO_DEVICENAME_SIZEOF                    ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_BUTTON  ,            0, "FAUXMOSAVEID"                , "Save and Update"         , NULL                                       , ACTION_SAVEFAUXMO                           ,   0  ,   0, NULL },
#endif
	{ PAGEID_PASS   , WSTYPE_TITLE   ,            0, NULL                          , "Miscellaneous"           , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_TEXT    ,            0, "HOSTNAMEID"                  , "Host Name:"              , (uint32_t*) &rlc.rlc_HostName              , HOSTNAME_SIZEOF                             ,   0  ,   0, NULL },
	{ PAGEID_PASS   , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },

	{ PAGEID_LOGIN  , WSTYPE_TITLE   ,            0, NULL                          , "Please Login:"           , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_LOGIN  , WSTYPE_PASSWORD,            0, "PASSWORD"                    , "Password:"               , NULL                                       , WEBSERVER_PASSWORD_SIZEOF                   ,   0  ,   0, NULL },
	{ PAGEID_LOGIN  , WSTYPE_HSPACE  , WSF_SR      , NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ PAGEID_LOGIN  , WSTYPE_BUTTON  ,            0, "LoginID"                     , "Submit"                  , NULL                                       , ACTION_NONE                                 ,   0  ,   0, NULL },
	{ PAGEID_LOGIN  , WSTYPE_SPACER  ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL },
	{ 0             , WSTYPE_INVALID ,            0, NULL                          , NULL                      , NULL                                       , 0                                           ,   0  ,   0, NULL }
};
/* \\\ HTML PAGE DATA */
/***************************************************************************/

/*
** HTML structural functions
*/

const char *DAT_PAGESPAN  = "<span style=\"margin:auto; width: auto; display: flex; justify-content: center; border:1px; background-color:lightblue\">";
const char *DAT_PAGESPAN2 = "<span style=\"margin:auto; width: auto; display: flex; justify-content: center; border:1px; background-color:#e1e1ff\">";

/*
** WebServer
*/

/* /// WebServer_AppendWiFiSignal()
**
*/

/***************************************************************************/

void WebServer_AppendWiFiSignal( void )
{

	float Signal = WiFi.RSSI();

	Signal = 90 / 40.0 * Signal + 212.5; // From Signal = 100% @ -50dBm and Signal = 10% @ -90dBm and y = mx + c
	if( Signal > 100 ) { 
		Signal = 100;
	}
	htmlpage += " ";
	htmlpage += String( Signal, 0 );
	htmlpage += "%";
}
/* \\\ WebServer_AppendWifiSignal() */
/* /// WebServer_AppendHeader()
**
*/

/***************************************************************************/

#define FUNC_CB "\
	function CB(oid) {\
		if( document.getElementById(oid).checked ) {\
			window.location.search   = '?' + oid + '=1';\
		} else {\
			window.location.search   = '?' + oid + '=0';\
		}\
	}"

#define FUNC_NM "\
	function NM(oid) {\
		window.location.search   = '?' + oid + '=' + document.getElementById(oid).value;\
	}"

#define FUNC_PW "\
function PW(oid) { \
	var op1 = document.getElementById(\"Pass1ID\");\
	var op2 = document.getElementById(\"Pass2ID\");\
	var oi  = document.getElementById(\"PassInfoID\");\
	var ost = document.getElementById(\"UpdateAndSaveID\");\
	var msg = \"\";\
	var dis = true;\
	var col = \"red\";\
	if( op1.value === op2.value ) {\
		if( op1.value.length < 6 ) {\
			msg = \"Password too short!\";\
		} else {\
			msg                      = \"Password is valid!\";\
			dis                      = false;\
			col                      = \"black\";\
			window.location.search   = \"?SECID=\" + op2.value;\
		}\
	} else {\
		msg = \"Passwords are not equal!\";\
	}\
	oi.innerHTML    = msg;\
	op1.style.color = col;\
	op2.style.color = col;\
	ost.disabled    = dis;\
}";

/***************************************************************************/

void WebServer_AppendHeader( bool refreshMode, uint32_t pageid ) 
{
uint32_t i;

	htmlpage  = "<!DOCTYPE html><html lang=\"en\">";
	htmlpage += "<head>";
	htmlpage += "<title>" WEBSITETITLE "</title>";
	htmlpage += "<meta charset=\"UTF-8\">";
	htmlpage += "<meta name=\"viewport\" content=\"width=device-width,initial-scale=1\">";
	if( refreshMode ) {
		htmlpage += "<meta http-equiv=\"refresh\" content=\"5\">"; // 5-secs refresh time, test needed to prevent auto updates repeating some commands
	}
/* CSS stuff */
	htmlpage += "<style>";
	htmlpage += "* {box-sizing: border-box;}";
	htmlpage += ".section {display: -webkit-flex; display: flex;}";
	htmlpage += ".topnav {overflow: hidden;background-color:lightcyan;}";
	htmlpage += ".topnav a {float:left;color:blue;text-align:center;padding:0.3em 0.3em;text-decoration:none;font-size:1.3em;}";
	htmlpage += ".topnav a:hover {background-color:deepskyblue;color:white;}";
	htmlpage += ".topnav a.active {background-color:lightblue;color:blue;}";

	htmlpage += ".body {font-family:Arial,Helvetica,sans-serif;font-size:14px;color:blue;background-color:#e1e1ff;text-align:center;}";

	htmlpage += "h2 {margin-top:0.3em;margin-bottom:0.3em;font-size:1.4em;}";  /* menu title */
	htmlpage += ".fsm {font-size:1.2em;padding:0;margin:0}";                   /* font size medium used in footer */
	htmlpage += ".fss {font-size:0.7em;padding:0;margin:0}";                   /* font size small in footer */
	htmlpage += "</style>";
/* java script section */
	htmlpage += "<script>";
	htmlpage += FUNC_CB;
	htmlpage += FUNC_NM;
	htmlpage += FUNC_PW;
	htmlpage += "</script>";

	htmlpage += "</head>";
/* body stuff */
	htmlpage += "<body>";
	htmlpage += "<section>";

	if( ( wifi_mode != WIFIMODE_STATE_APMODE ) && ( pageid != PAGEID_LOGIN ) ) {
		htmlpage += "<div class='topnav'>";
		htmlpage += DAT_PAGESPAN2;

		for( i = 0 ; i < PAGEIDDATA_TOPNAV_END ; i++ ) {
			htmlpage += "<a href='";
			htmlpage += TAB_PAGEIDDATA[ i ].p_URI;
			if( TAB_PAGEIDDATA[ i ].p_PageID == pageid ) {
				htmlpage += "' style='background-color:lightblue";
			}
			htmlpage += "'>";
			htmlpage += TAB_PAGEIDDATA[ i ].p_Name;
			htmlpage += "</a>";
		}
		htmlpage += "<span>";
		htmlpage += "</div>";
	}
}
/* \\\ WebServer_AppendHeader() */
/* /// WebServer_AppendFooter()
**
*/

/***************************************************************************/

void WebServer_AppendFooter( void )
{
	htmlpage += "</section>";
	htmlpage += "<footer>";
	htmlpage += DAT_PAGESPAN2;
	htmlpage += "<table><tr><td style=\"text-align:center\"><strong><p class=\"fsm\">" WEBSITEFOOTER " V" I2S(VERSION) "." I2S(REVISION) "</strong></p></td></tr>";
	htmlpage += "<tr><td style=\"text-align:center\"><p class=\"fss\"><i>Copyright &copy;" I2S(YEAR) "&nbsp;" AUTHORNAME "</i></td></tr></p>";
	htmlpage += "</table></span></footer></body></html>";
}
/* \\\ WebServer_AppendFooter() */
/* /// WebServer_Setup()
**
*/

/***************************************************************************/

void WebServer_Setup( void )
{

	debugex(( "\nWebServer_Setup()\n" ));

	server.close();

/* we want a user agent and a cookie */
	const char * headerkeys[]   = { "User-Agent","Cookie" };
	size_t       headerkeyssize = sizeof(headerkeys)/sizeof(char*);

/* we need to track these headers */
	server.collectHeaders( headerkeys, headerkeyssize );

	server.on( "/"        , [](void) { debug(( "server.on() \"/\"\n"       )); WebServer_HandleWebPage( PAGEID_STATUS              ); });
	server.on( "/status"  , [](void) { debug(( "server.on() \"/status\"\n" )); WebServer_HandleWebPage( PAGEID_STATUS              ); });
	server.on( "/login"   , [](void) { debug(( "server.on() \"/login\"\n"  )); WebServer_HandleWebPage( PAGEID_LOGIN               ); });
	server.on( "/visual"  , [](void) { debug(( "server.on() \"/visual\"\n" )); WebServer_HandleWebPage( PAGEID_VISUAL              ); });
	server.on( "/sound"   , [](void) { debug(( "server.on() \"/sound\"\n"  )); WebServer_HandleWebPage( PAGEID_SOUND               ); });
	server.on( "/global"  , [](void) { debug(( "server.on() \"/global\"\n" )); WebServer_HandleWebPage( PAGEID_GLOBAL              ); });
	server.on( "/misc"    , [](void) { debug(( "server.on() \"/misc\"\n"   )); WebServer_HandleWebPage( PAGEID_MISC                ); });
	server.on( "/wifi"    , [](void) { debug(( "server.on() \"/wifi\"\n"   )); WebServer_HandleWebPage( PAGEID_WIFI|PAGEID_PASS    ); });

	server.onNotFound( WebServer_HandleNotFound );

	server.begin();

#if ENABLE_DEBUG
	{
		float signal = WiFi.RSSI();
			signal = 90 / 40.0 * signal + 212.5; // From Signal = 100% @ -50dBm and Signal = 10% @ -90dBm and y = mx + c
			if( signal > 100 ) {
				signal = 100;
			};
			debug(( "Signal strength is %d%%\n", (uint8_t) signal ));
	}
#endif
}
/* \\\ WebServer_Setup() */
/* /// WebServer_HandleAction()
**
*/

/***************************************************************************/

void WebServer_HandleAction( void )
{
uint32_t action  = webserver_action; /* create a clone */

webserver_action = ACTION_NONE;   /* make sure we never execute this twice */
			debug(( "WebServer_HandleAction() - %d\n", webserver_action ));

			switch( action ) {
				case ACTION_SETTINGSLOAD:
					debug(( "WebServer_HandleAction() - ACTION_SETTINGSLOAD\n" ));
					Global_SettingsLoad();
					Global_HandleEvents( true );      /* update settings */
					break;
				case ACTION_SETTINGSSAVE:
					debug(( "WebServer_HandleAction() - ACTION_SETTINGSSAVE\n" ));
					Global_SettingsSave();
					break;
				case ACTION_SETTINGSRESET:
					Global_SettingsReset();
					Global_HandleEvents( true );      /* update settings */
					debug(( "WebServer_HandleAction() - ACTION_SETTINGSRESET\n" ));
					break;
				case ACTION_SWAPNETWORK:
					debug(( "WebServer_HandleAction() - ACTION_SETTINGSSWITCHWIFI\n" ));
					Wifi_SwapNetwork();
					break;
				case ACTION_SYSTEMREBOOT:
					debug(( "WebServer_HandleAction() - ACTION_SYSTEMREBOOT\n" ));
					ESP.restart();
					break;
				case ACTION_SOUNDTEST1:
					debug(( "WebServer_HandleAction() - ACTION_SOUNDTEST1\n" ));
					Sound_Play( SOUNDID_TEST_1, true );
					break;
				case ACTION_SOUNDTEST2:
					debug(( "WebServer_HandleAction() - ACTION_SOUNDTEST2\n" ));
					Sound_Play( SOUNDID_TEST_2, true );
					break;
				case ACTION_WIFISAVEREBOOT:
					debug(( "WebServer_HandleAction() - ACTION_WIFISAVEREBOOT\n" ));
/* save settings */
					Global_SettingsSave();
/* restart station mode and force a re-login */
					Wifi_SetupStationMode();
					break;
				case ACTION_CHANGEPASSWORD:
/* temppass contains the new password */
					strcpy( rlc.rlc_WebServerPassword, temppass );
/* this clears the retry text gadget */
					memset( temppass, 0, WEBSERVER_PASSWORD_SIZEOF ); 
/* save settings */
					Global_SettingsSave();
					debugex(( "WebServer_HandleAction() - New password is : '%s'\n", rlc.rlc_WebServerPassword ));
					break;
#if ENABLE_FAUXMO
				case ACTION_SAVEFAUXMO:
					Global_SettingsSave();
					Fauxmo_Rename();
					break;
#endif
		}
}
/* \\\ WebServer_HandleAction() */
/* /// WebServer_HandleWebPage()
**
*/

/***************************************************************************/

void WebServer_HandleWebPage( uint32_t pageid )
{
uint32_t res;

	debug(( "\nWebServer_HandleWebPage() - PageID: %s (%d)\n",
							( pageid ==  PAGEID_STATUS            ) ? "PAGEID_STATUS"           :
							( pageid ==  PAGEID_VISUAL            ) ? "PAGEID_VISUAL"           :
							( pageid ==  PAGEID_SOUND             ) ? "PAGEID_SOUND"            :
							( pageid ==  PAGEID_GLOBAL            ) ? "PAGEID_GLOBAL"           :
							( pageid ==  PAGEID_MISC              ) ? "PAGEID_MISC"             :
							( pageid ==  PAGEID_STATUS            ) ? "PAGEID_STATUS"           :
							( pageid ==  PAGEID_WIFI              ) ? "PAGEID_WIFI"             :
							( pageid ==  PAGEID_LOGIN             ) ? "PAGEID_LOGIN"            :
							( pageid ==  PAGEID_PASS              ) ? "PAGEID_PASS"             :
							( pageid == (PAGEID_WIFI|PAGEID_PASS) ) ? "PAGEID_WIFI|PAGEID_PASS" :
																	  "unknown" , pageid ));

#if !defined(DISABLE_LOGIN)             /* check if login is enabled */
/* login page for sure works without authentification */
	if( ( pageid == PAGEID_LOGIN ) ) {
		debug(( "WebServer_HandleWebPage() - WebServer_HandleLogin()\n" ));
		WebServer_HandleLogin();
		return;
	}
#endif

/* status works without authentification */
	if( ( pageid == PAGEID_STATUS ) ) {
		if( ( wifi_mode != WIFIMODE_STATE_APMODE ) ) {
			debug(( "WebServer_HandleWebPage() - No AP Mode: forwarding to status page\n" ));
			WebServer_CreateStatus(); 
			server.send( 200, "text/html", htmlpage );
			return;
		}
	}

#if !defined(DISABLE_LOGIN)             /* check if login is enabled */
/* Is the current user logged in? */
	if( !WebServer_IsAuthentified() ) {
		/* not logged in, so forward to login page */
		debug(( "WebServer_HandleWebPage() - Not authentified!: forwarding to login page\n" ));
		server.sendHeader( "Location","/login");
		server.sendHeader( "Cache-Control","no-cache");
		server.send( 307 );
		return;
	}
#endif

/* build the webpage based on the given pageid */
	WebServer_AppendHeader( false, pageid );
	htmlpage += DAT_PAGESPAN;
	res = HTML_CreateUIFromTable( pageid, &TAB_SETTINGS[0] );
	htmlpage += "</span>";
	WebServer_AppendFooter();

/* depending on the argument we redirect to our page without arguments */
	if( !res && server.args() ) {
		/* find the URI string for our web page */
		for( res = 0 ; TAB_PAGEIDDATA[ res ].p_PageID ; res++ ) {
			if( TAB_PAGEIDDATA[ res ].p_PageID == pageid ) {
				/* redirect */
				debug(( "WebServer_HandleWebPage() - post processing forward to %s page\n", TAB_PAGEIDDATA[ res ].p_URI ));
				server.sendHeader( "Location", TAB_PAGEIDDATA[ res ].p_URI );
				server.sendHeader( "Cache-Control","no-cache" );
				server.send( 307 );
				return;   /* avoid calling stuff below twice */
			}
		}
	} else {
		server.send( 200, "text/html", htmlpage ); 
	}
	Global_HandleEvents( true ); /* update settings */
	WebServer_HandleAction();
}
/* \\\ WebServer_HandleWebPage() */
/* /// WebServer_IsAuthentified()
**
** Check if header is present and correct
*/

/***************************************************************************/

bool WebServer_IsAuthentified()
{
	debugex(( "WebServer_IsAuthentified()\n" ));

	if( server.hasHeader( "Cookie" ) ) {
		String cookie = server.header( "Cookie" );
		debugex(( "WebServer_IsAuthentified() - Found cookie: %s\n", cookie.c_str() ));
		if( cookie.indexOf( WEBSERVER_COOKIENAME "=" + String( rlc.rlc_WebServerPassword ) ) != -1 ) {
			debugex(( "WebServer_IsAuthentified() - Authentification Successful\n" ));
			return( true );
		}
	}
	debugex(( "WebServer_IsAuthentified() - Authentification Failed\n" ));
	return( false );
}
/* \\\ */
/* /// WebServer_HandleLogin()
**
** login page, also called for disconnect
*/

/***************************************************************************/

void WebServer_HandleLogin()
{
	debugex(( "WebServer_HandleLogin()\n" ));

#if ENABLE_DEBUGEXTENDED
	if( server.hasHeader( "Cookie" ) ) {
		String cookie = server.header( "Cookie" );
		debugex(( "WebServer_HandleLogin() - Found cookie: '%s'\n", cookie.c_str() ));
	}
#endif
/* handle disconnect argument */
	if( server.hasArg( "DISCONNECT" ) ) {
		debug(( "WebServer_HandleLogin() - Disconnecting, resetting cookie\n" ));
		server.sendHeader( "Location","/login" );
		server.sendHeader( "Cache-Control","no-cache" );
		server.sendHeader( "Set-Cookie", WEBSERVER_COOKIENAME "=0" );
		server.send(307);
		return;
	}
/* handle password argument */
	if( server.hasArg( "PASSWORD" ) ) {
		debug(( "WebServer_HandleLogin() - PASSWORD found\n" ));
		debugex(( "WebServer_HandleLogin() - server.arg password is '%s'\n", server.arg( "PASSWORD" ).c_str() ));
		debugex(( "WebServer_HandleLogin() - Settings password is  '%s'\n", rlc.rlc_WebServerPassword ));
		if( server.arg( "PASSWORD" ) == rlc.rlc_WebServerPassword ) {
			debug(( "WebServer_HandleLogin() - PASSWORD is correct\n" ));
			server.sendHeader( "Location","/" );      /* redirect to our status page */
			server.sendHeader( "Cache-Control","no-cache" );
			server.sendHeader( "Set-Cookie",  WEBSERVER_COOKIENAME "=" + server.arg( "PASSWORD" ) );
			server.send(307);
			debug(( "WebServer_HandleLogin() - Cookie was set. Log in successful!\n" ));
			return;
		}
		debug(( "Log in Failed\n" ));
	}

/* build login page and send to server */
	WebServer_AppendHeader( false, PAGEID_LOGIN );
	htmlpage += DAT_PAGESPAN;
	HTML_CreateUIFromTable( PAGEID_LOGIN, &TAB_SETTINGS[0] );
	htmlpage += "</span>";
	WebServer_AppendFooter();
	server.send( 200, "text/html", htmlpage );
}
/* \\\ WebServer_HandleLogin() */
/* /// WebServer_HandleNotFound()
**
** no need authentification
*/

/***************************************************************************/

void WebServer_HandleNotFound()
{
/* build our 404 error page */
	htmlpage = "File Not Found\n\n";

	htmlpage += "URI: ";
	htmlpage += server.uri();
	htmlpage += "\nMethod: ";
	htmlpage += (server.method() == HTTP_GET)?"GET":"POST";
	htmlpage += "\nArguments: ";
	htmlpage += server.args();
	htmlpage += "\n";
	for( uint8_t i=0 ; i < server.args() ; i++ ) {
		htmlpage += " " + server.argName(i) + ": " + server.arg(i) + "\n";
	}
	server.send( 404, "text/plain", htmlpage );
}
/* \\\ WebServer_HandleNotFound() */
/* /// WebServer_CreateStatus()
**
*/

/***************************************************************************/

void WebServer_CreateStatus( void )
{
time_t now;
uint32_t i,c, r;
char buf[32];
const char *tableleft  = "<tr><td style=\"text-align:right\">";
const char *tableright = "</td><td style=\"text-align:left\">";
const char *tableclose = "</td></tr>";

	WebServer_AppendHeader( true, PAGEID_STATUS );

	now = time( nullptr );

	htmlpage += DAT_PAGESPAN;
	htmlpage += "<table border=\"0\">";

	for( int tabloop = 1 ; tabloop ; tabloop++ ) {
		htmlpage += tableleft;
		switch( tabloop ) {
			case 1:
				htmlpage += tableright;
				break;
			case 2:               /* version */
				htmlpage += "Version:";
				htmlpage += tableright;
				htmlpage += I2S(VERSION) "." I2S(REVISION) " - " I2S(DAY) "." I2S(MONTH) "." I2S(YEAR);
				break;
			case 3:               /* time */
				htmlpage += "System Time:";
				htmlpage += tableright;
				htmlpage += ctime( &now );
				break;
			case 4:               /* login */
				htmlpage += "Logged in:";
				htmlpage += tableright;
				htmlpage += ( WebServer_IsAuthentified() ? "Yes" : "No" );
				break;
			case 5:               /* Wifi signal strength */
				htmlpage += "Wifi signal strength:";
				htmlpage += tableright;
				WebServer_AppendWiFiSignal();
				break;
			case 6:               /* IP */
				htmlpage += "IP:";
				htmlpage += tableright;
				htmlpage += WiFi.localIP().toString();
				break;
			case 7:               /* Host name */
				htmlpage += "Host Name:";
				htmlpage += tableright;
				htmlpage += wifi_hostname;
				break;
			case 8:               /* Fauxmo name (Alexa) */
#if ENABLE_FAUXMO
				htmlpage += "Device Name:";
				htmlpage += tableright;
				htmlpage += rlc.rlc_FauxmoDeviceName;
				break;
#else 
				tabloop++; /* skip to next */
				// break; /* Fall through */
#endif
			case 9:               /* show global mode */
				htmlpage += "Global Mode:";
				htmlpage += tableright;
				htmlpage += TAB_CYCLE_GLOBALMODE[ rlrd.rlrd_GlobalMode ];
				break;
			case 10:              /* current power mode */
				htmlpage += "Power Mode:";
				htmlpage += tableright;
				htmlpage += ( rlc.rlc_PowerStateNightStart != rlc.rlc_PowerStateNightEnd ) ? ( ( rlrd.rlrd_Flags & RLRDF_STATUS_POWERSTATENIGHT ) ? "Night" : "Day" ) : "Day (DISABLED!)";
				break;
			case 11:              /* current display brightness mode */
				htmlpage += "Display Brightness Mode:";
				htmlpage += tableright;
				htmlpage += ( rlc.rlc_DisplayBrightnessNightStart != rlc.rlc_DisplayBrightnessNightEnd ) ? ( ( rlrd.rlrd_Flags & RLRDF_STATUS_DISPLAYBRIGHTNESSNIGHT ) ? "Night" : "Day" ) : "Day (DISABLED!)";
				break;
			case 12:              /* current sound volume mode */
				htmlpage += "Sound Volume Mode:";
				htmlpage += tableright;
				htmlpage += ( rlc.rlc_SoundVolumeNightStart != rlc.rlc_SoundVolumeNightEnd ) ? ( ( rlrd.rlrd_Flags & RLRDF_STATUS_SOUNDVOLUMENIGHT ) ? "Night" : "Day" ) : "Day (DISABLED!)";
				break;
			case 13:             /* inner ring LEDs */
				htmlpage += "Inner Ring LEDs (RGB):";
				htmlpage += tableright;
				sprintf( buf, "%d (%d)", HARDWARE_SEGMENT_1_LEDS, HARDWARE_SEGMENT_1_LEDS * 3 );
				htmlpage += buf;
				break;
			case 14:             /* outer ring LEDs */
				htmlpage += "Outer Ring LEDs (RGB):";
				htmlpage += tableright;
				sprintf( buf, "%d (%d)", HARDWARE_SEGMENT_2_LEDS, HARDWARE_SEGMENT_2_LEDS * 3 );
				htmlpage += buf;
				break;
			case 15:             /* total LEDs */
				htmlpage += "Total number of LEDs (RGB):";
				htmlpage += tableright;
				sprintf( buf, "%d (%d)", DISPLAY_TOTAL_LEDS, DISPLAY_TOTAL_LEDS * 3 );
				htmlpage += buf;
				break;
			case 16:             /* currently active LEDs */
				htmlpage += "LEDs active (RGB):";
				htmlpage += tableright;
				for( i = 0, c = 0, r = 0 ; i < DISPLAY_TOTAL_LEDS ; i++ ) {
					if(  ledfield[ i ].r || ledfield[ i ].g || ledfield[ i ].b ) { c++; }
					if( ledfield[ i ].r ) { r++; }
					if( ledfield[ i ].g ) { r++; }
					if( ledfield[ i ].b ) { r++; }
				}
				sprintf( buf, "%d (%d)", c, r );
				htmlpage += buf;
				break;
			case 17:             /* available power */
				htmlpage += "Available Power:";
				htmlpage += tableright;
				sprintf( buf, "%d mW (%d mA * %dV)", PSU_MILLIAMPS * PSU_VOLTAGE, PSU_MILLIAMPS, PSU_VOLTAGE );
				htmlpage += buf;
				break;
			case 18:             /* unscaled power */
				htmlpage += "Unscaled Power:";
				htmlpage += tableright;
				i = calculate_unscaled_power_mW( ledfield, DISPLAY_TOTAL_LEDS );
				sprintf( buf, "%d mW (%d mA * %dV)", i, i / PSU_VOLTAGE, PSU_VOLTAGE );
				htmlpage += buf;
				break;           /* last one, so fall through */
			default:
				tabloop = -1;    /* exit for() tabloop */
		}
		htmlpage += tableclose;
	}   /* for( tabloop ) */
	htmlpage += "</table>";
	htmlpage += "</span>";

	WebServer_AppendFooter();
}
/* \\\ WebServer_CreateStatus() */

/***************************************************************************/
